// SPDX-License-Identifier: BSD-3-Clause
// Necula Mihail 323 CAa 2024 - 2025

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/epoll.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/sendfile.h>
#include <sys/eventfd.h>
#include <libaio.h>
#include <errno.h>

#include "aws.h"
#include "utils/util.h"
#include "utils/debug.h"
#include "utils/sock_util.h"
#include "utils/w_epoll.h"

/* server socket file descriptor */
static int listenfd;

/* epoll file descriptor */
static int epollfd;


/*****
 * This function is used by http parser.
 *****/
static int aws_on_path_cb(http_parser *p, const char *buf, size_t len)
{
	struct connection *conn = (struct connection *)p->data;

	conn->request_path[0] = '.';
	memcpy(conn->request_path + 1, buf, len);
	conn->request_path[len + 1] = '\0';
	conn->have_path = 1;

	return 0;
}

/*****
 * Allocate memory for a connection and initialize the fields of the structure.
 *
 * @param sockf the fd through which the server and the client comunicate
 * @return the created connection
 *****/
struct connection *connection_create(int sockfd)
{
	struct connection *conn;

	conn = (struct connection *)malloc(sizeof(struct connection));
	DIE(!conn, "malloc() failed\n");

	conn->fd = -1;
	conn->filename[0] = '\0';

	conn->eventfd = -1;
	conn->sockfd = sockfd;

	memset(&conn->ctx, 0, sizeof(conn->ctx));
	memset(&conn->iocb, 0, sizeof(conn->iocb));
	conn->piocb[0] = &conn->iocb;
	conn->file_size = 0;

	conn->recv_buffer[0] = '\0';
	conn->recv_len = 0;

	conn->send_buffer[0] = '\0';
	conn->send_len = 0;
	conn->send_pos = 0;
	conn->file_pos = 0;
	conn->async_read_len = 0;

	conn->have_path = 0;
	conn->request_path[0] = '\0';
	conn->res_type = RESOURCE_TYPE_NONE;
	conn->state = STATE_INITIAL;

	return conn;
}

/*****
 * Close a connection with a client.
 *****/
void connection_remove(struct connection *conn)
{
	char client_addr[101];
	int rc;

	rc = get_peer_address(conn->sockfd, client_addr, 100);
	DIE(rc, "get_peer_address() failed\n");

	rc = epoll_ctl(epollfd, EPOLL_CTL_DEL, conn->sockfd, NULL);
	DIE(rc, "epol_ctl() failed\n");

	close(conn->sockfd);
	free(conn);

	dlog(LOG_INFO, "Server closed connection with: %s.\n", client_addr);
}

/*****
 * Handle a new connection request on the server socket.
 *****/
void handle_new_connection(void)
{
	struct sockaddr_in client_addr;
	socklen_t size_addr = sizeof(client_addr);
	int rc, clientfd, flags;
	struct connection *conn;

	/* Accept new connection. */
	clientfd = accept(listenfd, (struct sockaddr *) &client_addr, &size_addr);
	DIE(clientfd < 0, "accept() failed\n");
	dlog(LOG_INFO, "Server accepted connection from: %s:%d\n",
		inet_ntoa(client_addr.sin_addr), ntohs(client_addr.sin_port));

	/* Set socket to be non-blocking. */
	flags = fcntl(clientfd, F_GETFL, 0);
	DIE(flags == -1, "fcntl() failed\n");
	rc = fcntl(clientfd, F_SETFL, flags | O_NONBLOCK);
	DIE(rc == -1, "fcntl() failed\n");

	/* Instantiate new connection handler. */
	conn = connection_create(clientfd);

	/* Add socket to epoll. */
	rc = w_epoll_add_ptr_in(epollfd, clientfd, conn);
	DIE(rc, "w_epoll_add_ptr_in() failed\n");

	/* Initialize HTTP_REQUEST parser. */
	http_parser_init(&conn->request_parser, HTTP_REQUEST);
	conn->request_parser.data = (void *)conn;
}

/*****
 * Receive message on socket.
 * Store message in recv_buffer in struct connection.
 *****/
int receive_data(struct connection *conn)
{
	char client_addr[101], *buffer;
	int recv_bytes;

	get_peer_address(conn->sockfd, client_addr, 100);
	conn->state = STATE_RECEIVING_DATA;

	buffer = conn->recv_buffer + conn->recv_len;
	recv_bytes = recv(conn->sockfd, buffer, BUFSIZ - conn->recv_len, 0);

	if (recv_bytes <= 0) {
		dlog(LOG_INFO, "Server can;t comunicat with: %s.\n", client_addr);
		connection_remove(conn);
		return -1;
	}

	conn->recv_len += recv_bytes;
	return 0;
}

/*****
 * Analysing the requested pathname, decides if the file is of type STATIC or DYNAMIC.
 *
 * @param conn contains the path of the requested file
 *****/
void connection_set_resource_type(struct connection *conn)
{
	if (conn->request_path[2] == 's')
		conn->res_type = RESOURCE_TYPE_STATIC;
	else if (conn->request_path[2] == 'd')
		conn->res_type = RESOURCE_TYPE_DYNAMIC;
}

/*****
 * Use this function to parse the request header from clients.
 *****/
void parse_header(struct connection *conn)
{
	/* TODO: Parse the HTTP header and extract the file path. */
	/* Use mostly null settings except for on_path callback. */
	http_parser_settings settings_on_path = {
		.on_message_begin = 0,
		.on_header_field = 0,
		.on_header_value = 0,
		.on_path = aws_on_path_cb,
		.on_url = 0,
		.on_fragment = 0,
		.on_query_string = 0,
		.on_body = 0,
		.on_headers_complete = 0,
		.on_message_complete = 0
	};

	http_parser_execute(&conn->request_parser, &settings_on_path,
		conn->recv_buffer, conn->recv_len);
	connection_set_resource_type(conn);
}

/*****
 * Open the request file by client.
 *****/
void connection_open_file(struct connection *conn)
{
	struct stat st;

	conn->fd = open(conn->request_path, O_RDONLY);

	if (conn->fd < 0) {
		strcpy(conn->send_buffer, HTTP_404);
		conn->send_len = strlen(HTTP_404);
	} else {
		strcpy(conn->send_buffer, HTTP_200);
		conn->send_len = strlen(HTTP_200);

		stat(conn->request_path, &st);
		conn->file_size = st.st_size;
		conn->file_pos = 0;
	}
}

/*****
 * Send data from a buffer to a client.
 *
 * @param conn conatains the buffer (send_buffer) and the client (scokfd)
 *****/
int connection_send_buffer(struct connection *conn)
{
	char client_addr[101];
	int rc, bytes_sent;

	rc = get_peer_address(conn->sockfd, client_addr, 100);
	DIE(rc == -1, "get_peer_address() failed\n");

	bytes_sent = send(conn->sockfd, conn->send_buffer, conn->send_len, 0);

	if (bytes_sent <= 0) {
		dlog(LOG_INFO, "Server can't comunicate with: %s.\n",
			client_addr);
		connection_remove(conn);
		return -1;
	}

	conn->send_len -= bytes_sent;
	memmove(conn->send_buffer, conn->send_buffer + bytes_sent, conn->send_len + 1);

	return 0;
}

/*****
 * Sends a file to a client using zero-coppy.
 *****/
int connection_send_static(struct connection *conn)
{
	if (!conn->file_size) {
		connection_remove(conn);
		return -1;
	}

	char client_addr[101];
	int rc, bytes_sent;

	rc = get_peer_address(conn->sockfd, client_addr, 100);
	DIE(rc == -1, "get_peer_address() failed\n");

	bytes_sent = sendfile(conn->sockfd, conn->fd, &conn->file_pos, conn->file_size);

	if (bytes_sent <= 0) {
		dlog(LOG_INFO, "Server can't comunicate with: %s.\n",
			client_addr);
		connection_remove(conn);
		return -1;
	}

	conn->file_size -= bytes_sent;
	return 0;
}

/*****
 * Starts asynchronous operation (read from file).
 * Uses io_submit(2) & friends for reading data asynchronously.
 *****/
void connection_start_async_io(struct connection *conn)
{
	int rc;

	/* Prepare the io control block. */
	conn->async_read_len = 16 * BUFSIZ < conn->file_size ? 16 * BUFSIZ : conn->file_size;
	memset(&conn->iocb, 0, sizeof(conn->iocb));
	io_prep_pread(&conn->iocb, conn->fd, conn->send_buffer, conn->async_read_len, conn->file_pos);

	/* Associate the operation with a fd. */
	conn->eventfd = eventfd(0, 0);
	DIE(conn->eventfd == -1, "eventfd() failed\n");
	io_set_eventfd(&conn->iocb, conn->eventfd);

	/* We won't send something to client, until this operation is done. */
	rc = w_epoll_remove_ptr(epollfd, conn->sockfd, conn);
	DIE(rc == -1, "w_epoll_remove_ptr() failed\n");

	/* We need to know when the operation is done. */
	rc = w_epoll_add_ptr_in(epollfd, conn->eventfd, conn);
	DIE(rc == -1, "w_epoll_add_ptr_in() failed\n");

	/* Setup the context. */
	memset(&conn->ctx, 0, sizeof(io_context_t));
	rc = io_setup(1, &conn->ctx);
	DIE(rc, "io_setup() failed\n");

	/* Start the operation. */
	rc = io_submit(conn->ctx, 1, conn->piocb);
	DIE(rc < 0, "io_submit() failed\n");
}

/*****
 * Complete asynchronous operation. Prepare socket for sending.
 *****/
void connection_complete_async_io(struct connection *conn)
{
	int rc;

	rc = w_epoll_remove_ptr(epollfd, conn->eventfd, conn);
	DIE(rc == -1, "w_epoll_remove_ptr() failed\n");

	rc = w_epoll_add_ptr_out(epollfd, conn->sockfd, conn);
	DIE(rc == -1, "w_epoll_add_ptr_out() failed\n");

	/* Why i can not put this at begining? */
	rc = io_destroy(conn->ctx);
	DIE(rc, "io_destroy() failed\n");

	rc = close(conn->eventfd);
	DIE(rc, "close() failed\n");

	conn->eventfd = -1;
	conn->send_len = conn->async_read_len;
	conn->async_read_len = 0;
	conn->file_pos += conn->send_len;
	conn->file_size -= conn->send_len;
}

/*****
 * Read data asynchronously and put it in connection send buffer,
 * which is further sent to client.
 *****/
int connection_send_dynamic(struct connection *conn)
{
	char client_addr[101];
	int rc;

	rc = get_peer_address(conn->sockfd, client_addr, 100);
	DIE(rc == -1, "get_peer_address() failed\n");

	if (!conn->send_len && conn->file_size) {
		connection_start_async_io(conn);
		return 0;
	}

	connection_send_buffer(conn);
	return 0;
}

/*****
 * Send as much data as possible from the connection send buffer / connection file.
 *****/
int connection_send_data(struct connection *conn)
{
	char client_addr[101];
	int rc;

	rc = get_peer_address(conn->sockfd, client_addr, 100);
	DIE(rc == -1, "get_peer_address() failed\n");

	if (conn->state == STATE_SENDING_HEADER) {
		rc = connection_send_buffer(conn);

		if (!conn->send_len) {
			dlog(LOG_INFO, "Server sent reply header to: %s.\n", client_addr);
			conn->state = STATE_HEADER_SENT;

			if (!conn->file_size)
				connection_remove(conn);
			else
				conn->state = STATE_SENDING_DATA;
		}
	}

	if (conn->state == STATE_SENDING_DATA) {
		if (conn->res_type == RESOURCE_TYPE_STATIC)
			rc = connection_send_static(conn);
		else
			rc = connection_send_dynamic(conn);

		if (!conn->send_len && !conn->file_size) {
			dlog(LOG_INFO, "Server sent a file to: %s.\n", client_addr);
			connection_remove(conn);
		}
	}

	return rc;
}


/*****
 * Handles the request of a client.
 * Receives data from him and sends data to him.
 *****/
void handle_client(uint32_t event, struct connection *conn)
{
	int rc;
	char *buffer_end;

	if (event & EPOLLIN) {
		rc = receive_data(conn);

		if (rc == -1)
			return;
		buffer_end = conn->recv_buffer + conn->recv_len - 4;
		if (strncmp(buffer_end, HTTP_HEADER_END, 4))
			return;
		conn->state = STATE_REQUEST_RECEIVED;

		parse_header(conn);
		connection_open_file(conn);

		rc = w_epoll_update_ptr_out(epollfd, conn->sockfd, conn);
		DIE(rc == -1, "w_epoll_update_ptr_out() failed\n");
		conn->state = STATE_SENDING_HEADER;
	}

	if (event & EPOLLOUT) {
		rc = connection_send_data(conn);
		DIE(rc == -1, "connection_send_data() failed\n");
	}
}

int main(void)
{
	int rc;

	/* Initialize multiplexing. */
	epollfd = w_epoll_create();
	DIE(epollfd == -1, "w_epol_create() failed\n");

	/* Create server socket. */
	listenfd = tcp_create_listener(AWS_LISTEN_PORT, DEFAULT_LISTEN_BACKLOG);
	DIE(listenfd < 0, "tcp_create_listener() failed\n");

	/* Add server socket to epoll object*/
	rc = w_epoll_add_fd_in(epollfd, listenfd);
	DIE(rc, "w_epol_add_fd_in() failed\n");

	dlog(LOG_INFO, "Server waiting for connections on port %d\n", AWS_LISTEN_PORT);

	/* server main loop */
	while (1) {
		struct epoll_event recv_ev;

		/* Wait for events. */
		rc = epoll_wait(epollfd, &recv_ev, 1, -1);
		DIE(rc <= 0, "epoll_wait() failed\n");

		/* Switch event type:
		 *   - new connection requests (on server socket)
		 *   - socket communication (on connection sockets)
		 */
		struct connection *conn = (struct connection *)recv_ev.data.ptr;

		if (recv_ev.data.fd == listenfd)
			handle_new_connection();
		else if (conn->eventfd >= 0)
			connection_complete_async_io(conn);
		else
			handle_client(recv_ev.events, conn);
	}

	close(listenfd);
	close(epollfd);

	return 0;
}
